@@ -27,20 +27,23 @@ from django.contrib.auth.models import AnonymousUser
 from django.core.exceptions import ImproperlyConfigured
 from django.test import Client, TestCase, override_settings
 from django.test.client import RequestFactory
-from django.urls import reverse
+from django.urls import reverse, reverse_lazy
 from django.utils.encoding import force_text
 from djangosaml2 import views
 from djangosaml2.cache import OutstandingQueriesCache
 from djangosaml2.conf import get_config
 from djangosaml2.middleware import SamlSessionMiddleware
 from djangosaml2.tests import conf
-from djangosaml2.utils import (get_idp_sso_supported_bindings,
+from djangosaml2.utils import (get_fallback_login_redirect_url,
+                               get_idp_sso_supported_bindings,
                                get_session_id_from_saml2,
                                get_subject_id_from_saml2,
                                saml2_from_httpredirect_request)
 from djangosaml2.views import (EchoAttributesView, finish_logout)
 from saml2.config import SPConfig
-from saml2.s_utils import decode_base64_and_inflate, deflate_and_base64_encode
+from saml2.s_utils import (decode_base64_and_inflate,
+                           deflate_and_base64_encode,
+                           UnknownSystemEntity)
 
 from .auth_response import auth_response
 from .utils import SAMLPostFormParser
@@ -74,6 +77,23 @@ class UtilsTests(TestCase):
         with self.assertRaisesMessage(ImproperlyConfigured, 'Module "djangosaml2.tests" does not define a "nonexisting_function" attribute/class'):
             get_config('djangosaml2.tests.nonexisting_function')
 
+    @override_settings(LOGIN_REDIRECT_URL='/accounts/profile/')
+    def test_get_fallback_login_redirect_url(self):
+        self.assertEqual(get_fallback_login_redirect_url(), '/accounts/profile/')
+
+        with override_settings():
+            del settings.LOGIN_REDIRECT_URL
+            # Neither LOGIN_REDIRECT_URL nor ACS_DEFAULT_REDIRECT_URL is configured
+            self.assertEqual(get_fallback_login_redirect_url(), '/')
+
+        with override_settings(ACS_DEFAULT_REDIRECT_URL='testprofiles:dashboard'):
+            # ACS_DEFAULT_REDIRECT_URL is configured, so it is used (and resolved)
+            self.assertEqual(get_fallback_login_redirect_url(), '/dashboard/')
+
+        with override_settings(ACS_DEFAULT_REDIRECT_URL=reverse_lazy('testprofiles:dashboard')):
+            # Lazy urls are resolved
+            self.assertEqual(get_fallback_login_redirect_url(), '/dashboard/')
+
 
 class SAML2Tests(TestCase):
 
@@ -114,8 +134,9 @@ class SAML2Tests(TestCase):
             idp_hosts=['idp.example.com'],
             metadata_file='remote_metadata_one_idp.xml',
         )
-        self.assertEqual(get_idp_sso_supported_bindings(
-            idp_entity_id='random'), [])
+
+        with self.assertRaises(UnknownSystemEntity):
+            get_idp_sso_supported_bindings(idp_entity_id='random')
 
     def test_get_idp_sso_supported_bindings_no_idps(self):
         settings.SAML_CONFIG = conf.create_conf(
@@ -155,7 +176,7 @@ class SAML2Tests(TestCase):
     def test_login_evil_redirect(self):
         """
         Make sure that if we give an URL other than our own host as the next
-        parameter, it is replaced with the default LOGIN_REDIRECT_URL.
+        parameter, it is replaced with the fallback login redirect url.
         """
 
         # monkey patch SAML configuration
@@ -164,17 +185,30 @@ class SAML2Tests(TestCase):
             idp_hosts=['idp.example.com'],
             metadata_file='remote_metadata_one_idp.xml',
         )
-        response = self.client.get(
-            reverse('saml2_login') + '?next=http://evil.com')
-        url = urlparse(response['Location'])
-        params = parse_qs(url.query)
 
-        self.assertEqual(params['RelayState'], [settings.LOGIN_REDIRECT_URL, ])
+        for redirect_url in ['/dashboard/', 'testprofiles:dashboard']:
+            with self.subTest(LOGIN_REDIRECT_URL=redirect_url):
+                with override_settings(LOGIN_REDIRECT_URL=redirect_url):
+                    response = self.client.get(
+                        reverse('saml2_login') + '?next=http://evil.com')
+                    url = urlparse(response['Location'])
+                    params = parse_qs(url.query)
+
+                    self.assertEqual(params['RelayState'], ['/dashboard/'])
+
+            with self.subTest(ACS_DEFAULT_REDIRECT_URL=redirect_url):
+                with override_settings(ACS_DEFAULT_REDIRECT_URL=redirect_url):
+                    response = self.client.get(
+                        reverse('saml2_login') + '?next=http://evil.com')
+                    url = urlparse(response['Location'])
+                    params = parse_qs(url.query)
+
+                    self.assertEqual(params['RelayState'], ['/dashboard/'])
 
     def test_no_redirect(self):
         """
         Make sure that if we give an empty path as the next parameter,
-        it is replaced with the default LOGIN_REDIRECT_URL.
+        it is replaced with the fallback login redirect url.
         """
 
         # monkey patch SAML configuration
@@ -183,11 +217,58 @@ class SAML2Tests(TestCase):
             idp_hosts=['idp.example.com'],
             metadata_file='remote_metadata_one_idp.xml',
         )
-        response = self.client.get(reverse('saml2_login') + '?next=')
-        url = urlparse(response['Location'])
-        params = parse_qs(url.query)
 
-        self.assertEqual(params['RelayState'], [settings.LOGIN_REDIRECT_URL, ])
+        for redirect_url in ['/dashboard/', 'testprofiles:dashboard']:
+            with self.subTest(LOGIN_REDIRECT_URL=redirect_url):
+                with override_settings(LOGIN_REDIRECT_URL=redirect_url):
+                    response = self.client.get(reverse('saml2_login') + '?next=')
+                    url = urlparse(response['Location'])
+                    params = parse_qs(url.query)
+
+                    self.assertEqual(params['RelayState'], ['/dashboard/'])
+
+            with self.subTest(ACS_DEFAULT_REDIRECT_URL=redirect_url):
+                with override_settings(ACS_DEFAULT_REDIRECT_URL=redirect_url):
+                    response = self.client.get(reverse('saml2_login') + '?next=')
+                    url = urlparse(response['Location'])
+                    params = parse_qs(url.query)
+
+                    self.assertEqual(params['RelayState'], ['/dashboard/'])
+
+    @override_settings(SAML_IGNORE_AUTHENTICATED_USERS_ON_LOGIN=True)
+    def test_login_already_logged(self):
+        self.client.force_login(User.objects.create(username='user', password='pass'))
+
+        for redirect_url in ['/dashboard/', 'testprofiles:dashboard']:
+            with self.subTest(LOGIN_REDIRECT_URL=redirect_url):
+                with override_settings(LOGIN_REDIRECT_URL=redirect_url):
+                    with self.subTest('no next url'):
+                        response = self.client.get(reverse('saml2_login'))
+                        self.assertRedirects(response, '/dashboard/')
+
+                    with self.subTest('evil next url'):
+                        response = self.client.get(reverse('saml2_login') + '?next=http://evil.com')
+                        self.assertRedirects(response, '/dashboard/')
+
+            with self.subTest(ACS_DEFAULT_REDIRECT_URL=redirect_url):
+                with override_settings(ACS_DEFAULT_REDIRECT_URL=redirect_url):
+                    with self.subTest('no next url'):
+                        response = self.client.get(reverse('saml2_login'))
+                        self.assertRedirects(response, '/dashboard/')
+
+                    with self.subTest('evil next url'):
+                        response = self.client.get(reverse('saml2_login') + '?next=http://evil.com')
+                        self.assertRedirects(response, '/dashboard/')
+
+    def test_unknown_idp(self):
+        # monkey patch SAML configuration
+        settings.SAML_CONFIG = conf.create_conf(
+            sp_host='sp.example.com',
+            metadata_file='remote_metadata_three_idps.xml',
+        )
+
+        response = self.client.get(reverse('saml2_login')+'?idp=https://unknown.org')
+        self.assertEqual(response.status_code, 403)
 
     def test_login_one_idp(self):
         # monkey patch SAML configuration
@@ -264,6 +345,7 @@ class SAML2Tests(TestCase):
         self.assertIn('AuthnRequest xmlns', decode_base64_and_inflate(
             saml_request).decode('utf-8'))
 
+    @override_settings(ACS_DEFAULT_REDIRECT_URL='testprofiles:dashboard')
     def test_assertion_consumer_service(self):
         # Get initial number of users
         initial_user_count = User.objects.count()
@@ -312,14 +394,36 @@ class SAML2Tests(TestCase):
             'SAMLResponse': self.b64_for_post(saml_response),
             'RelayState': came_from,
         })
-        self.assertEqual(response.status_code, 302)
-        location = response['Location']
 
-        url = urlparse(location)
-        # as the RelayState is empty we have redirect to LOGIN_REDIRECT_URL
-        self.assertEqual(url.path, settings.LOGIN_REDIRECT_URL)
+        # as the RelayState is empty we have redirect to ACS_DEFAULT_REDIRECT_URL
+        self.assertRedirects(response, '/dashboard/')
         self.assertEqual(force_text(new_user.id), client.session[SESSION_KEY])
 
+    @override_settings(ACS_DEFAULT_REDIRECT_URL='testprofiles:dashboard')
+    def test_assertion_consumer_service_default_relay_state(self):
+        settings.SAML_CONFIG = conf.create_conf(
+            sp_host='sp.example.com',
+            idp_hosts=['idp.example.com'],
+            metadata_file='remote_metadata_one_idp.xml',
+        )
+
+        new_user = User.objects.create(username='teacher', password='not-used')
+
+        response = self.client.get(reverse('saml2_login'))
+        saml2_req = saml2_from_httpredirect_request(response.url)
+        session_id = get_session_id_from_saml2(saml2_req)
+
+        saml_response = auth_response(session_id, 'teacher')
+        self.add_outstanding_query(session_id, '/')
+        response = self.client.post(reverse('saml2_acs'), {
+            'SAMLResponse': self.b64_for_post(saml_response),
+        })
+        self.assertEqual(response.status_code, 302)
+
+        # The RelayState is missing, redirect to ACS_DEFAULT_REDIRECT_URL
+        self.assertRedirects(response, '/dashboard/')
+        self.assertEqual(force_text(new_user.id), self.client.session[SESSION_KEY])
+
     def test_assertion_consumer_service_already_logged_in_allowed(self):
         self.client.force_login(User.objects.create(
             username='user', password='pass'))
@@ -616,6 +720,17 @@ class SAML2Tests(TestCase):
         _args, kwargs = prepare_for_auth_mock.call_args
         self.assertIn('sigalg', kwargs)
 
+    @override_settings(SAML2_DISCO_URL="https://that-ds.org/ds")
+    def test_discovery_service(self):
+        settings.SAML_CONFIG = conf.create_conf(
+            sp_host='sp.example.com',
+            idp_hosts=['idp.example.com'],
+            metadata_file='remote_metadata_three_idps.xml',
+        )
+
+        response = self.client.get(reverse('saml2_login'))
+        self.assertEqual(response.status_code, 302)
+        self.assertIn("https://that-ds.org/ds", response.url)
 
 def test_config_loader(request):
     config = SPConfig()
@@ -746,6 +861,5 @@ class MiddlewareTests(SessionEnabledTestCase):
             cookie = response.cookies[saml_session_name]
 
             self.assertIsNotNone(cookie['expires'])
-
             self.assertNotEqual(cookie['expires'], '')
             self.assertNotEqual(cookie['max-age'], '')
