@@ -21,6 +21,7 @@ from typing import Optional
 from django.conf import settings
 from django.core.exceptions import ImproperlyConfigured
 from django.http import HttpResponse, HttpResponseRedirect
+from django.shortcuts import resolve_url
 from django.utils.http import is_safe_url
 from saml2.config import SPConfig
 from saml2.s_utils import UnknownSystemEntity
@@ -68,8 +69,9 @@ def get_idp_sso_supported_bindings(idp_entity_id: Optional[str] = None, config:
     try:
         return list(meta.service(idp_entity_id, 'idpsso_descriptor', 'single_sign_on_service').keys())
     except UnknownSystemEntity:
-        return []
-
+        raise UnknownSystemEntity
+    except Exception as e:
+        logger.error(f"get_idp_sso_supported_bindings failed with: {e}")
 
 def get_location(http_info):
     """Extract the redirect URL from a pysaml2 http_info object"""
@@ -80,6 +82,11 @@ def get_location(http_info):
         return http_info['url']
 
 
+def get_fallback_login_redirect_url():
+    login_redirect_url = get_custom_setting('LOGIN_REDIRECT_URL', '/')
+    return resolve_url(get_custom_setting('ACS_DEFAULT_REDIRECT_URL', login_redirect_url))
+
+
 def validate_referral_url(request, url):
     # Ensure the user-originating redirection url is safe.
     # By setting SAML_ALLOWED_HOSTS in settings.py the user may provide a list of "allowed"
@@ -90,7 +97,7 @@ def validate_referral_url(request, url):
         getattr(settings, 'SAML_ALLOWED_HOSTS', [request.get_host()]))
 
     if not is_safe_url(url=url, allowed_hosts=saml_allowed_hosts):
-        return settings.LOGIN_REDIRECT_URL
+        return get_fallback_login_redirect_url()
     return url
 
 
@@ -104,23 +111,25 @@ def saml2_from_httpredirect_request(url):
 
 
 def get_session_id_from_saml2(saml2_xml):
-    saml2_xml = saml2_xml.encode() if isinstance(saml2_xml, str) else saml2_xml
-    return re.findall(b'ID="([a-z0-9\-]*)"', saml2_xml, re.I)[0].decode()
+    saml2_xml = saml2_xml.decode() if isinstance(saml2_xml, bytes) else saml2_xml
+    return re.findall(r'ID="([a-z0-9\-]*)"', saml2_xml, re.I)[0]
 
 
 def get_subject_id_from_saml2(saml2_xml):
     saml2_xml = saml2_xml if isinstance(saml2_xml, str) else saml2_xml.decode()
     re.findall('">([a-z0-9]+)</saml:NameID>', saml2_xml)[0]
 
-def add_param_in_url(url:str, param_key:str, param_value:str):
+
+def add_param_in_url(url: str, param_key: str, param_value: str):
     params = list(url.split('?'))
     params.append(f'{param_key}={param_value}')
-    new_url = params[0] + '?' +''.join(params[1:])
+    new_url = params[0] + '?' + ''.join(params[1:])
     return new_url
 
+
 def add_idp_hinting(request, http_response) -> bool:
     idphin_param = getattr(settings, 'SAML2_IDPHINT_PARAM', 'idphint')
-    params = urllib.parse.urlencode(request.GET)
+    urllib.parse.urlencode(request.GET)
 
     if idphin_param not in request.GET.keys():
         return False
