@@ -43,17 +43,18 @@ from saml2.response import (RequestVersionTooLow,
                             UnsolicitedResponse)
 from saml2.s_utils import UnsupportedBinding
 from saml2.saml import SCM_BEARER
-from saml2.samlp import AuthnRequest
+from saml2.saml import AuthnContextClassRef
+from saml2.samlp import RequestedAuthnContext
+from saml2.samlp import AuthnRequest, IDPEntry, IDPList, Scoping
 from saml2.sigver import MissingKey
 from saml2.validate import ResponseLifetimeExceed, ToEarly
-from saml2.xmldsig import (  # support for SHA1 is required by spec
-    SIG_RSA_SHA1, SIG_RSA_SHA256)
 
 from .cache import IdentityCache, OutstandingQueriesCache, StateCache
 from .conf import get_config
 from .exceptions import IdPConfigurationMissing
 from .overrides import Saml2Client
 from .utils import (add_idp_hinting, available_idps, get_custom_setting,
+                    get_fallback_login_redirect_url,
                     get_idp_sso_supported_bindings, get_location,
                     validate_referral_url)
 
@@ -118,7 +119,7 @@ class LoginView(SPConfigMixin, View):
             If the user is already logged in (and if allowed), he will redirect to there immediately.
         '''
 
-        next_path = settings.LOGIN_REDIRECT_URL
+        next_path = get_fallback_login_redirect_url()
         if 'next' in request.GET:
             next_path = request.GET['next']
         elif 'RelayState' in request.GET:
@@ -127,6 +128,48 @@ class LoginView(SPConfigMixin, View):
         next_path = validate_referral_url(request, next_path)
         return next_path
 
+    def unknown_idp(self, request, idp):
+        msg = (f'Error: IdP EntityID {idp} was not found in metadata')
+        logger.error(msg)
+        return HttpResponse(
+            msg.format('Please contact technical support.'), status=403
+        )
+
+    def load_sso_kwargs_scoping(self, sso_kwargs):
+        """ Performs IdP Scoping if scoping param is present. """
+        idp_scoping_param = self.request.GET.get('scoping', None)
+        if idp_scoping_param:
+            idp_scoping = Scoping()
+            idp_scoping.idp_list = IDPList()
+            idp_scoping.idp_list.idp_entry.append(
+                IDPEntry(provider_id = idp_scoping_param)
+            )
+            sso_kwargs['scoping'] = idp_scoping
+
+    def load_sso_kwargs_authn_context(self, sso_kwargs):
+        # this would work when https://github.com/IdentityPython/pysaml2/pull/807
+        ac = getattr(self.conf, '_sp_requested_authn_context', {})
+
+        # this works even without https://github.com/IdentityPython/pysaml2/pull/807
+        # hopefully to be removed soon !
+        if not ac:
+            scs = getattr(
+                settings, 'SAML_CONFIG', {}
+            ).get('service', {}).get('sp', {})
+            ac = scs.get('requested_authn_context', {})
+        # end transitional things to be removed soon !
+
+        if ac:
+            sso_kwargs["requested_authn_context"] = RequestedAuthnContext(
+                    authn_context_class_ref=[
+                        AuthnContextClassRef(ac['authn_context_class_ref']),
+                    ],
+                    comparison = ac.get('comparison', "minimum"),
+                )
+
+    def load_sso_kwargs(self, sso_kwargs):
+        """ Inherit me if you want to put your desidered things in sso_kwargs """
+
     def get(self, request, *args, **kwargs):
         logger.debug('Login process started')
         next_path = self.get_next_path(request)
@@ -151,15 +194,18 @@ class LoginView(SPConfigMixin, View):
 
         try:
             conf = self.get_sp_config(request)
-        except SourceNotFound as excp:
-            msg = ('Error, IdP EntityID was not found in metadata: {}')
-            logger.exception(msg.format(excp))
-            return HttpResponse(msg.format('Please contact technical support.'), status=500)
+        except SourceNotFound as excp:  # pragma: no cover
+            # this is deprecated and it's here only for the doubts that something
+            # would happen the day after I'll remove it! :)
+            return self.unknown_idp(request, idp='unknown')
 
         # is a embedded wayf or DiscoveryService needed?
         configured_idps = available_idps(conf)
         selected_idp = request.GET.get('idp', None)
 
+        self.conf = conf
+        sso_kwargs = {}
+
         # Do we have a Discovery Service?
         if not selected_idp:
             discovery_service = getattr(settings, 'SAML2_DISCO_URL', None)
@@ -186,19 +232,23 @@ class LoginView(SPConfigMixin, View):
                 })
 
         # is the first one, otherwise next logger message will print None
-        if not configured_idps:
+        if not configured_idps: # pragma: no cover
             raise IdPConfigurationMissing(
-                ('IdP configuration is missing or its metadata is expired.'))
+                ('IdP is missing or its metadata is expired.'))
         if selected_idp is None:
             selected_idp = list(configured_idps.keys())[0]
 
         # choose a binding to try first
-        binding = getattr(settings, 'SAML_DEFAULT_BINDING', saml2.BINDING_HTTP_POST)
+        binding = getattr(settings, 'SAML_DEFAULT_BINDING',
+                          saml2.BINDING_HTTP_POST)
         logger.debug(f'Trying binding {binding} for IDP {selected_idp}')
 
         # ensure our selected binding is supported by the IDP
-        supported_bindings = get_idp_sso_supported_bindings(
-            selected_idp, config=conf)
+        try:
+            supported_bindings = get_idp_sso_supported_bindings(
+                selected_idp, config=conf)
+        except saml2.s_utils.UnknownSystemEntity:
+            return self.unknown_idp(request, selected_idp)
 
         if binding not in supported_bindings:
             logger.debug(
@@ -211,33 +261,31 @@ class LoginView(SPConfigMixin, View):
                     f'trying {saml2.BINDING_HTTP_REDIRECT}',
                 )
                 binding = saml2.BINDING_HTTP_REDIRECT
-            else:
+            else: # pragma: no cover
                 logger.warning(
                     f'IDP {selected_idp} does not support {binding} '
                     f'trying {saml2.BINDING_HTTP_POST}',
                 )
                 binding = saml2.BINDING_HTTP_POST
             # if switched binding still not supported, give up
-            if binding not in supported_bindings:
+            if binding not in supported_bindings: # pragma: no cover
                 raise UnsupportedBinding(
                     f'IDP {selected_idp} does not support '
-                    f'{saml2.BINDING_HTTP_POST} and {saml2.BINDING_HTTP_REDIRECT}'
+                    f'{saml2.BINDING_HTTP_POST} or {saml2.BINDING_HTTP_REDIRECT}'
                 )
 
         client = Saml2Client(conf)
-        http_response = None
 
         # SSO options
         sign_requests = getattr(conf, '_sp_authn_requests_signed', False)
-        sso_kwargs = {}
         if sign_requests:
-            sso_kwargs["sigalg"] = settings.SAML_CONFIG['service']['sp']\
-                                           .get('signing_algorithm',
-                                                saml2.xmldsig.SIG_RSA_SHA256)
-            sso_kwargs["digest_alg"] = settings.SAML_CONFIG['service']['sp']\
-                                           .get('digest_algorithm',
-                                                saml2.xmldsig.DIGEST_SHA256)
-
+            sso_kwargs["sigalg"] = getattr(conf, '_sp_signing_algorithm',
+                                           saml2.xmldsig.SIG_RSA_SHA256
+            )
+            sso_kwargs["digest_alg"] = getattr(conf,
+                                               '_sp_digest_algorithm',
+                                               saml2.xmldsig.DIGEST_SHA256
+            )
         # pysaml needs a string otherwise: "cannot serialize True (type bool)"
         if getattr(conf, '_sp_force_authn', False):
             sso_kwargs['force_authn'] = "true"
@@ -247,8 +295,19 @@ class LoginView(SPConfigMixin, View):
         # custom nsprefixes
         sso_kwargs['nsprefix'] = get_namespace_prefixes()
 
+
+        # Enrich sso_kwargs ...
+        # idp scoping
+        self.load_sso_kwargs_scoping(sso_kwargs)
+        # authn context
+        self.load_sso_kwargs_authn_context(sso_kwargs)
+        # other customization to be inherited
+        self.load_sso_kwargs(sso_kwargs)
+
         logger.debug(f'Redirecting user to the IdP via {binding} binding.')
         _msg = 'Unable to know which IdP to use'
+        http_response = None
+
         if binding == saml2.BINDING_HTTP_REDIRECT:
             try:
                 session_id, result = client.prepare_for_authenticate(
@@ -260,6 +319,7 @@ class LoginView(SPConfigMixin, View):
                 return HttpResponse(_msg)
             else:
                 http_response = HttpResponseRedirect(get_location(result))
+
         elif binding == saml2.BINDING_HTTP_POST:
             if self.post_binding_form_template:
                 # get request XML to build our own html based on the template
@@ -268,10 +328,12 @@ class LoginView(SPConfigMixin, View):
                 except TypeError as e:
                     logger.error(f'{_msg}: {e}')
                     return HttpResponse(_msg)
+
                 session_id, request_xml = client.create_authn_request(
                     location,
                     binding=binding,
-                    **sso_kwargs)
+                    **sso_kwargs
+                )
                 try:
                     if isinstance(request_xml, AuthnRequest):
                         # request_xml will be an instance of AuthnRequest if the message is not signed
@@ -287,14 +349,16 @@ class LoginView(SPConfigMixin, View):
                         },
                     })
                 except TemplateDoesNotExist as e:
-                    logger.error(f'TemplateDoesNotExist: {e}')
+                    logger.error(
+                        f'TemplateDoesNotExist: [{self.post_binding_form_template}] - {e}'
+                    )
 
             if not http_response:
                 # use the html provided by pysaml2 if no template was specified or it doesn't exist
                 try:
                     session_id, result = client.prepare_for_authenticate(
                         entityid=selected_idp, relay_state=next_path,
-                        binding=binding)
+                        binding=binding, **sso_kwargs)
                 except TypeError as e:
                     _msg = f"Can't prepare the authentication for {selected_idp}"
                     logger.error(f'{_msg}: {e}')
@@ -373,7 +437,8 @@ class AssertionConsumerServiceView(SPConfigMixin, View):
         except ResponseLifetimeExceed as e:
             _exception = e
             logger.info(
-                ("SAML Assertion is no longer valid. Possibly caused by network delay or replay attack."), exc_info=True)
+                ("SAML Assertion is no longer valid. Possibly caused "
+                 "by network delay or replay attack."), exc_info=True)
         except SignatureError as e:
             _exception = e
             logger.info("Invalid or malformed SAML Assertion.", exc_info=True)
@@ -428,7 +493,8 @@ class AssertionConsumerServiceView(SPConfigMixin, View):
         for sc in assertion.subject.subject_confirmation:
             if sc.method == SCM_BEARER:
                 assertion_not_on_or_after = sc.subject_confirmation_data.not_on_or_after
-                assertion_info = {'assertion_id': assertion.id, 'not_on_or_after': assertion_not_on_or_after}
+                assertion_info = {'assertion_id': assertion.id,
+                                  'not_on_or_after': assertion_not_on_or_after}
                 break
 
         if callable(attribute_mapping):
@@ -472,9 +538,7 @@ class AssertionConsumerServiceView(SPConfigMixin, View):
     def build_relay_state(self) -> str:
         """ The relay state is a URL used to redirect the user to the view where they came from.
         """
-        login_redirect_url = get_custom_setting('LOGIN_REDIRECT_URL', '/')
-        default_relay_state = get_custom_setting(
-            'ACS_DEFAULT_REDIRECT_URL', login_redirect_url)
+        default_relay_state = get_fallback_login_redirect_url()
         relay_state = self.request.POST.get('RelayState', default_relay_state)
         relay_state = self.customize_relay_state(relay_state)
         if not relay_state:
@@ -600,10 +664,14 @@ class LogoutView(SPConfigMixin, View):
     logout_error_template = 'djangosaml2/logout_error.html'
 
     def get(self, request, *args, **kwargs):
-        return self.do_logout_service(request, request.GET, saml2.BINDING_HTTP_REDIRECT, *args, **kwargs)
+        return self.do_logout_service(
+            request, request.GET, saml2.BINDING_HTTP_REDIRECT, *args, **kwargs
+        )
 
     def post(self, request, *args, **kwargs):
-        return self.do_logout_service(request, request.POST, saml2.BINDING_HTTP_POST, *args, **kwargs)
+        return self.do_logout_service(
+            request, request.POST, saml2.BINDING_HTTP_POST, *args, **kwargs
+        )
 
     def do_logout_service(self, request, data, binding):
         logger.debug('Logout service started')
